"""Defines the commands for executing the module directly."""

from __future__ import annotations

import argparse
import asyncio
import logging
import mimetypes
import pathlib
import sys
from typing import TYPE_CHECKING

from . import retry

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable


def _upload(args: argparse.Namespace) -> int:
    """Implement the "upload" command.

    Returns the exit status for the program.
    """
    metadata: dict[str, bytes | None] = {
        "filename": pathlib.Path(args.file).name.encode()
    }

    if mime_type := mimetypes.guess_type(args.file)[0]:
        metadata["mime_type"] = mime_type.encode()

    for meta in args.metadata:
        kv = meta.split("=", maxsplit=1)
        metadata[kv[0]] = kv[1].encode() if (len(kv) == 2) else None  # noqa: PLR2004

    try:
        with pathlib.Path(args.file).open("rb") as file:
            if location := asyncio.run(retry.upload(args.endpoint, file, metadata)):
                print(str(location))  # noqa: T201
                return 0
    except KeyboardInterrupt:  # pragma: no cover
        pass
    except Exception as e:  # noqa: BLE001
        logging.error("Unable to upload file: %s", e)  # noqa: LOG015 TRY400

    return 1


def _metadata(args: argparse.Namespace) -> int:
    """Implement the "metadata" command.

    Returns the exit status for the program.
    """
    try:
        metadata = asyncio.run(retry.metadata(args.location))
        if metadata is None:
            return 1

        for k, v in metadata.items():
            if v is None:
                print(f"{k}")  # noqa: T201
            else:
                value = repr(v)[2:][:-1]
                print(f"{k}: {value}")  # noqa: T201
    except KeyboardInterrupt:  # pragma: no cover
        return 1

    return 0


def main(argv: list[str]) -> int:
    """Entrypoint function for when the module is executed directly.

    :param argv: The command line arguments to the program.
    :return: Exit status for the program.
    """
    interpreter = pathlib.Path(sys.executable).name if sys.executable else "python3"

    parser = argparse.ArgumentParser(prog=f"{interpreter} -m aiotus")
    parser.add_argument("--debug", action="store_true", help="log debug messages")
    subparsers = parser.add_subparsers()

    parser_upload = subparsers.add_parser(
        "upload",
        help="Upload a file to a tus (tus.io) server.",
    )
    parser_upload.add_argument(
        "--metadata",
        action="append",
        default=[],
        help="additional metadata to upload ('key[=value]')",
    )
    parser_upload.add_argument(
        "endpoint", type=str, help="creation URL of the tus server"
    )
    parser_upload.add_argument("file", type=str, help="file to upload")
    parser_upload.set_defaults(func=_upload)

    parser_metadata = subparsers.add_parser(
        "metadata",
        help="Query the metadata of a file on a tus (tus.io) server.",
    )
    parser_metadata.add_argument(
        "location", type=str, help="file location on the tus server"
    )
    parser_metadata.set_defaults(func=_metadata)

    args = parser.parse_args(args=argv)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if not hasattr(args, "func"):
        sys.stderr.write("No command specified.\n\n")
        parser.print_help(file=sys.stderr)
        return 1

    func: Callable[[argparse.Namespace], int] = args.func
    return func(args)



"""Test uploading to a server behind a TLS proxy."""

from __future__ import annotations

import shutil
import ssl
from typing import TYPE_CHECKING

import aiohttp
import pytest

import aiotus

if TYPE_CHECKING:  # pragma: no cover
    import io

    from . import conftest


@pytest.mark.skipif(shutil.which("nginx") is None, reason="nginx not found")
class TestTLS:
    async def test_upload_fail(
        self, nginx_proxy: conftest.TusServer, memory_file: io.BytesIO
    ) -> None:
        """Test failed upload to a TLS server."""

        # Make sure we actually use encryption, access via plain
        # HTTP shall fail.
        http_url = nginx_proxy.url.with_scheme("http")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiohttp.ClientResponseError) as excinfo:
                await aiotus.creation.create(session, http_url, memory_file, {})
            assert excinfo.value.status == 400

        # As we use a self-signed certificate, the connection will fail.
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiohttp.ClientConnectorCertificateError) as excinfo:
                await aiotus.creation.create(session, nginx_proxy.url, memory_file, {})

            messages = [
                "certificate verify failed: self signed certificate",
                "certificate verify failed: self-signed certificate",
            ]
            assert any(m in str(excinfo.value) for m in messages)

        # The retrying upload function returns 'None', as the upload fails.
        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)
        location = await aiotus.upload(nginx_proxy.url, memory_file, config=config)
        assert location is None

        # Uploading with TLS verification disabled shall work.
        async with aiohttp.ClientSession() as session:
            await aiotus.creation.create(
                session, nginx_proxy.url, memory_file, {}, ssl=False
            )

        # With TLS verification disabled the creation works.
        config = aiotus.RetryConfiguration(ssl=False)
        location = await aiotus.upload(nginx_proxy.url, memory_file, config=config)
        assert location is not None

        # Fetching the metadata over HTTP fails.
        http_location = location.with_scheme("http")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiohttp.ClientResponseError) as excinfo:
                await aiotus.core.metadata(session, http_location)
            assert "Bad Request" in str(excinfo.value)

    async def test_upload_functional(
        self, nginx_proxy: conftest.TusServer, memory_file: io.BytesIO
    ) -> None:
        """Test creation on a TLS server."""

        ssl_ctx = ssl.create_default_context(cafile=nginx_proxy.certificate)
        async with aiohttp.ClientSession() as session:
            await aiotus.creation.create(
                session, nginx_proxy.url, memory_file, {}, ssl=ssl_ctx
            )

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001, ssl=ssl_ctx)
        location = await aiotus.upload(nginx_proxy.url, memory_file, config=config)
        assert location is not None



from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import io
import logging
import math
import os.path
import tempfile
from typing import TYPE_CHECKING

import aiohttp
import pytest
import pytest_asyncio
import yarl

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import AsyncGenerator, Callable, Mapping
    from typing import Any

    import pytest_aiohttp

logger = logging.getLogger(__name__)


@dataclasses.dataclass(kw_only=True)
class MockTusServer:
    # Number of times the respective handlers will fail.
    retries_create: int
    retries_options: int
    retries_head: int
    retries_upload: int

    # URLs to create and upload.
    create_endpoint: yarl.URL
    upload_endpoint: yarl.URL

    # The uploaded data will be accumulated here.
    data: bytearray | None

    # Metadata included in the creation will be placed here.
    metadata: str | None

    # Complete HTTP headers used in the last head/post request.
    head_headers: Mapping[str, str] | None
    post_headers: Mapping[str, str] | None

    # Drop some bytes while uploading, but don't return an error.
    drop_upload: bool

    # The aiohttp test server object.
    server: aiohttp.test_utils.TestServer


@pytest_asyncio.fixture
async def tus_server(  # noqa: C901 PLR0915
    aiohttp_server: pytest_aiohttp.AiohttpServer,
) -> MockTusServer:
    """Return a fake tus server that can consume a single file."""

    async def handler_create(request: aiohttp.web.Request) -> aiohttp.web.Response:
        server: MockTusServer = request.app["state"]["server"]

        server.retries_create -= 1
        if server.retries_create > 0:
            raise aiohttp.web.HTTPInternalServerError

        if "Upload-Metadata" in request.headers:
            server.metadata = request.headers["Upload-Metadata"]

        server.post_headers = request.headers

        # "Create" the upload.
        server.data = bytearray()

        headers = {"Location": str(server.upload_endpoint)}
        raise aiohttp.web.HTTPCreated(headers=headers)

    async def handler_options(request: aiohttp.web.Request) -> aiohttp.web.Response:
        server: MockTusServer = request.app["state"]["server"]

        server.retries_options -= 1
        if server.retries_options > 0:
            raise aiohttp.web.HTTPInternalServerError

        headers = {
            "Tus-Resumable": "1.0.0",
            "Tus-Version": "1.0.0",
            "Tus-Extension": "creation",
        }

        raise aiohttp.web.HTTPNoContent(headers=headers)

    async def handler_head(request: aiohttp.web.Request) -> aiohttp.web.Response:
        server: MockTusServer = request.app["state"]["server"]

        server.retries_head -= 1
        if server.retries_head > 0:
            raise aiohttp.web.HTTPInternalServerError

        server.head_headers = request.headers

        if server.data is None:
            raise aiohttp.web.HTTPNotFound

        headers = {"Upload-Offset": str(len(server.data))}
        if server.metadata is not None:
            headers["Upload-Metadata"] = server.metadata
        raise aiohttp.web.HTTPOk(headers=headers)

    async def handler_upload(request: aiohttp.web.Request) -> aiohttp.web.Response:
        server: MockTusServer = request.app["state"]["server"]

        body = await request.read()

        if server.data is None:
            raise aiohttp.web.HTTPUnprocessableEntity

        if int(request.headers["Upload-Offset"]) != len(server.data):
            raise aiohttp.web.HTTPConflict

        server.retries_upload -= 1
        if server.retries_upload > 0:
            # Pretend we did only receive half of the data before an error happend.
            server.data.extend(body[: len(body) // 2])
            raise aiohttp.web.HTTPInternalServerError

        if server.drop_upload:
            # Simulate the situation where the server can only store a subset
            # of the data that was uploaded. (In contrast to the error that we simulate
            # above we do not return an HTTP error status code.)
            body = body[: math.ceil(len(body) / 2.0)]

        server.data.extend(body)
        headers = {"Tus-Resumable": "1.0.0", "Upload-Offset": str(len(server.data))}
        raise aiohttp.web.HTTPNoContent(headers=headers)

    upload_name = "1234abcdefgh"

    app = aiohttp.web.Application()
    app["state"] = {}
    app.router.add_route("POST", "/files", handler_create)
    app.router.add_route("OPTIONS", "/files", handler_options)
    app.router.add_route("HEAD", "/files/" + upload_name, handler_head)
    app.router.add_route("PATCH", "/files/" + upload_name, handler_upload)

    http_server = await aiohttp_server(app)
    create_endpoint = http_server.make_url("/files")
    upload_endpoint = create_endpoint / upload_name

    server = MockTusServer(
        retries_create=0,
        retries_options=0,
        retries_head=0,
        retries_upload=0,
        create_endpoint=create_endpoint,
        upload_endpoint=upload_endpoint,
        data=None,
        metadata=None,
        head_headers=None,
        post_headers=None,
        drop_upload=False,
        server=http_server,
    )

    app["state"]["server"] = server

    return server


@pytest.fixture
def memory_file() -> io.BytesIO:
    """Dummy data to use during tests."""
    return io.BytesIO(b"\x00\x01\x02\x03")


class EOFBytesIO:
    """A wrapper around 'io.BytesIO' that never returns data."""

    def __init__(self, b: io.BytesIO) -> None:
        self._b = b

    def seek(
        self,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> int:
        return self._b.seek(*args, **kwargs)

    def read(
        self,
        *args: Any,  # noqa: ANN401 ARG002
        **kwargs: Any,  # noqa: ANN401 ARG002
    ) -> bytes:
        return b""


@pytest.fixture
def eof_memory_file(memory_file: io.BytesIO) -> io.BytesIO:
    # The two functions implemented by 'BinaryIO' are the only
    # ones aiotus uses, so the type error can be ignored.
    return EOFBytesIO(memory_file)  # type: ignore[return-value]


@dataclasses.dataclass
class TusServer:
    # The URL where the server is listening on.
    url: yarl.URL

    # The path of the certificate file of the server, if TLS is used.
    certificate: str | None = None


@contextlib.asynccontextmanager
async def start_process(
    program: str, args: list[str], *, cwd: str
) -> AsyncGenerator[None]:
    """Start a subprocess."""

    proc = await asyncio.create_subprocess_exec(program, *args, cwd=cwd)
    logger.info("process '%s' started", program)

    try:
        await asyncio.sleep(1.0)
        yield
    finally:
        logger.info("terminating process '%s'...", program)
        proc.terminate()
        await proc.communicate()
        logger.info("process '%s' terminated", program)


@pytest_asyncio.fixture(loop_scope="module")
async def tusd(
    pytestconfig: pytest.Config, unused_tcp_port_factory: Callable[[], int]
) -> AsyncGenerator[TusServer]:
    """Start the tusd (tus.io reference implementation) and yield the upload URL.

    Assumes that the tusd executable is located in the pytest rootdir.
    """
    host = "0.0.0.0"  # noqa: S104
    port = unused_tcp_port_factory()
    basepath = "/files/"

    with tempfile.TemporaryDirectory() as d:
        async with start_process(
            str(pytestconfig.rootpath / "tusd"),
            ["-host", host, "-port", str(port), "-base-path", basepath],
            cwd=d,
        ):
            logger.info("tusd started, listening on port %s", port)
            yield TusServer(yarl.URL(f"http://{host}:{port}{basepath}"))


# Template for the nginx configuration file. Stripped-down from
# https://github.com/tus/tusd/blob/master/examples/nginx.conf
_nginx_conf = """
    pid nginx.pid;
    daemon off;
    error_log /dev/null;

    events {{
    }}

    http {{
        client_body_temp_path ./client_body;
        proxy_temp_path ./proxy;
        fastcgi_temp_path ./fastcgi;
        uwsgi_temp_path ./uwsgi;
        scgi_temp_path ./scgi;

        access_log off;

        server {{
            listen {port} ssl;
            server_name _;
            ssl_certificate {crt};
            ssl_certificate_key {key};

            location / {{
                # Forward incoming requests to local tusd instance
                proxy_pass {tusd_url};

                # Disable request and response buffering
                proxy_request_buffering off;
                proxy_buffering off;
                proxy_http_version 1.1;

                # Add X-Forwarded-* headers
                proxy_set_header X-Forwarded-Host $hostname;
                proxy_set_header X-Forwarded-Proto $scheme;

                proxy_set_header Upgrade $http_upgrade;
                proxy_set_header Connection "upgrade";
                client_max_body_size 0;
            }}
        }}
    }}
"""


@pytest_asyncio.fixture(loop_scope="module")
async def nginx_proxy(
    tusd: TusServer, unused_tcp_port_factory: Callable[[], int]
) -> AsyncGenerator[TusServer]:
    """Start an nginx proxy in front of tusd that does TLS termination."""
    test_dir = os.path.dirname(os.path.abspath(__file__))
    certificate = os.path.join(test_dir, "selfsigned.crt")
    port = unused_tcp_port_factory()

    fmt = {
        "crt": certificate,
        "key": os.path.join(test_dir, "nginx.key"),
        "port": port,
        "tusd_url": str(tusd.url),
    }
    conf = _nginx_conf.format(**fmt)

    with tempfile.TemporaryDirectory() as d:
        conf_file = os.path.join(d, "nginx.conf")
        with open(conf_file, "w") as f:
            f.write(conf)

        async with start_process("nginx", ["-p", d, "-c", conf_file], cwd=d):
            logger.info("nginx started, listening on port %s", port)
            yield TusServer(
                yarl.URL(f"https://localhost:{port}"),
                certificate,
            )



"""Implementation of the core protocol.

The
`core tus protocol <https://tus.io/protocols/resumable-upload.html#core-protocol>`_
defines how the data upload is handled.
"""

from __future__ import annotations

import asyncio
import base64
import dataclasses
import io
from typing import TYPE_CHECKING

from . import common
from .log import logger

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping
    from typing import BinaryIO

    import aiohttp
    import multidict
    import yarl


@dataclasses.dataclass
class ServerConfiguration:
    """Class to hold the server's configuration."""

    protocol_versions: list[str]
    """
    List of protocol versions supported by the server, sorted by the server's
    preference.
    """

    max_size: int | None
    """
    The maximum allowed file size (in bytes), if reported by the server.
    """

    protocol_extensions: list[str]
    """
    The protocol extensions supported by the server.
    """


def _parse_positive_integer_header(
    headers: multidict.CIMultiDictProxy[str], header_name: str
) -> int:
    """Convert a HTTP header into a positive integer value.

    Raises a ProtocolError if the conversion is not posible.
    """
    if header_name not in headers:
        msg = f'HTTP header "{header_name}" not included in server response.'
        raise common.ProtocolError(msg)

    header_value = headers[header_name]

    try:
        if (result := int(header_value)) < 0:
            raise RuntimeError  # noqa: TRY301
    except Exception as e:
        msg = (
            f'Unable to convert "{header_name}" header '
            f'"{header_value}" to a positive integer.'
        )
        raise common.ProtocolError(msg) from e

    return result


async def offset(
    session: aiohttp.ClientSession,
    location: yarl.URL,
    ssl: common.SSLArgument = True,  # noqa: FBT002
    headers: Mapping[str, str] | None = None,
) -> int:
    """Get the number of uploaded bytes.

    :param session: HTTP session to use for connections.
    :param location: The upload endpoint to query.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: The number of bytes that are already on the server.
    """
    tus_headers = dict(headers or {})
    tus_headers["Tus-Resumable"] = common.TUS_PROTOCOL_VERSION

    logger.debug(f'Getting offset of "{location}"...')
    async with await session.head(location, headers=tus_headers, ssl=ssl) as response:
        response.raise_for_status()

        return _parse_positive_integer_header(response.headers, "Upload-Offset")


def _parse_metadata(header: str) -> common.Metadata:
    """Split and decode the input into a metadata dictionary."""
    if not (header := header.strip()):
        return {}

    md: dict[str, bytes | None] = {}
    for pair in header.split(","):
        kv = pair.split()
        if len(kv) == 1:
            md[kv[0]] = None
        elif len(kv) == 2:  # noqa: PLR2004
            md[kv[0]] = base64.b64decode(kv[1], validate=True)
        else:
            msg = "Key/Value pair consists of more than two elements."
            raise ValueError(msg)

    return md


async def metadata(
    session: aiohttp.ClientSession,
    location: yarl.URL,
    ssl: common.SSLArgument = True,  # noqa: FBT002
    headers: Mapping[str, str] | None = None,
) -> common.Metadata:
    """Get the metadata associated with an upload.

    See :data:`aiotus.Metadata` for details on how metadata is handled in the
    tus protocol.

    :param session: HTTP session to use for connections.
    :param location: The upload endpoint to query.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: The metadata of the upload.
    :raises common.ProtocolError: When the server does not comply to the tus protocol.
    """
    tus_headers = dict(headers or {})
    tus_headers["Tus-Resumable"] = common.TUS_PROTOCOL_VERSION

    logger.debug(f'Getting metadata of "{location}"...')
    async with await session.head(location, headers=tus_headers, ssl=ssl) as response:
        response.raise_for_status()

        if "Upload-Metadata" not in response.headers:
            return {}

        try:
            return _parse_metadata(response.headers["Upload-Metadata"])
        except Exception as e:
            msg = f"Unable to parse metadata: {e}"
            raise common.ProtocolError(msg) from e


async def upload_buffer(  # noqa: PLR0913
    session: aiohttp.ClientSession,
    location: yarl.URL,
    buffer: BinaryIO,
    ssl: common.SSLArgument = True,  # noqa: FBT002
    chunksize: int = 4 * 1024 * 1024,
    headers: Mapping[str, str] | None = None,
) -> None:
    """Upload data to the server.

    :param session: HTTP session to use for connections.
    :param location: The endpoint to upload to.
    :param buffer: The data to upload.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param chunksize: The size of individual chunks to upload at a time.
    :param headers: Optional headers used in the request.
    :raises common.ProtocolError: When the server does not comply to the tus protocol.
    :raises RuntimeError: When reading of the file fails.
    """
    total_size = await asyncio.to_thread(buffer.seek, 0, io.SEEK_END)

    # The position in the file where we currently read from.
    current_read_offset = -1

    # We ask the server for the number of bytes it already has for the upload. This
    # makes it possible to use this function also for resuming aborted uploads.
    current_server_offset = await offset(session, location, ssl=ssl, headers=headers)

    logger.debug(
        f'Resuming upload of "{location}" at offset {current_server_offset}..."'
    )

    while True:
        if current_server_offset == total_size:
            # Done, the whole file is on the server.
            msg = "Complete buffer uploaded."
            logger.info(msg)
            break

        if current_server_offset > total_size:
            # The offset that the server expects next does not exist.
            msg = "Server offset too big."
            raise common.ProtocolError(msg)

        if current_read_offset != current_server_offset:
            # Seek to the offset that the server expects next.
            current_read_offset = current_server_offset
            await asyncio.to_thread(buffer.seek, current_read_offset, io.SEEK_SET)

        if not (chunk := await asyncio.to_thread(buffer.read, chunksize)):
            # If the checks above are correct, we should never get here.
            msg = "Buffer returned unexpected EOF."
            raise RuntimeError(msg)

        current_read_offset += len(chunk)

        tus_headers = dict(headers or {})
        tus_headers.update(
            {
                "Tus-Resumable": common.TUS_PROTOCOL_VERSION,
                "Upload-Offset": str(current_server_offset),
                "Content-Length": str(len(chunk)),
                "Content-Type": "application/offset+octet-stream",
            }
        )

        logger.debug(f'Uploading {len(chunk)} bytes to "{location}"...')
        async with await session.patch(
            location, headers=tus_headers, data=chunk, ssl=ssl
        ) as response:
            response.raise_for_status()

            # Safe the value of the current offset on the server-side, at the beginning
            # of this loop are checks to see if it is valid.
            current_server_offset = _parse_positive_integer_header(
                response.headers, "Upload-Offset"
            )


async def configuration(
    session: aiohttp.ClientSession,
    url: yarl.URL,
    ssl: common.SSLArgument = True,  # noqa: FBT002
    headers: Mapping[str, str] | None = None,
) -> ServerConfiguration:
    """Get the server's configuration.

    :param session: HTTP session to use for connections.
    :param url: The creation endpoint of the server.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: An object describing the server's configuration.
    :raises common.ProtocolError: When the server does not comply to the tus protocol.
    """
    logger.debug("Querying server configuration...")
    async with await session.options(url, headers=headers, ssl=ssl) as response:
        response.raise_for_status()

        if "Tus-Version" not in response.headers:
            msg = '"Tus-Version" header not present.'
            raise common.ProtocolError(msg)

        versions = response.headers["Tus-Version"].split(",")
        versions = [v.strip() for v in versions]

        max_size = None
        if "Tus-Max-Size" in response.headers:
            max_size = _parse_positive_integer_header(response.headers, "Tus-Max-Size")

        extensions = []
        if "Tus-Extension" in response.headers:
            extensions = response.headers["Tus-Extension"].split(",")
            extensions = [e.strip() for e in extensions]

        return ServerConfiguration(versions, max_size, extensions)



"""Common constants and classes used in the aiotus package."""

from __future__ import annotations

import ssl
from collections.abc import Mapping
from typing import Final, TypeAlias

import aiohttp

# The version of the tus protocol we implement.
TUS_PROTOCOL_VERSION: Final = "1.0.0"

SSLArgument: TypeAlias = bool | ssl.SSLContext | aiohttp.Fingerprint

Metadata: TypeAlias = Mapping[str, bytes | None]


class ProtocolError(Exception):
    """Server response did not follow the tus protocol."""



"""Implementation of the creation extension.

The
`creation extension <https://tus.io/protocols/resumable-upload.html#creation>`_
defines how to reserve space on the server for uploading data to.
"""

from __future__ import annotations

import asyncio
import base64
import io
from typing import TYPE_CHECKING

import yarl

from . import common
from .log import logger

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping
    from typing import BinaryIO

    import aiohttp


def _check_metadata_keys(metadata: common.Metadata) -> None:
    """Check if the metadata keys are valid.

    Raises a 'ValueError' exception if a key is invalid.
    """
    for k in metadata:
        if not k.isascii():
            msg = "Metadata keys must only contain ASCII characters."
            raise ValueError(msg)

        if " " in k:
            msg = "Metadata keys must not contain spaces."
            raise ValueError(msg)

        if "," in k:
            msg = "Metadata keys must not contain commas."
            raise ValueError(msg)


def encode_metadata(metadata: common.Metadata) -> str:
    """Encode the metadata to the value of the metadata header.

    :param metadata: The metadata to encode.
    :return: The value for the "Upload-Metadata" header.
    """
    _check_metadata_keys(metadata)

    def encode_value(value: bytes | None) -> str:
        if value is None:
            return ""

        encoded_bytes = base64.b64encode(value)
        encoded_string = encoded_bytes.decode()
        return " " + encoded_string

    pairs = [f"{k}{encode_value(v)}" for k, v in metadata.items()]
    return ",".join(pairs)


async def create(  # noqa: PLR0913
    session: aiohttp.ClientSession,
    url: yarl.URL,
    file: BinaryIO | None,
    metadata: common.Metadata,
    ssl: common.SSLArgument = True,  # noqa: FBT002
    headers: Mapping[str, str] | None = None,
) -> yarl.URL:
    """Create an upload.

    :param session: HTTP session to use for connections.
    :param url: The creation endpoint of the server.
    :param file: The file object to upload.
        Used to determine the length of data to be uploaded. If not given, the
        corresponding HTTP header is not included in the request.
    :param metadata: Additional metadata for the upload.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: The URL to upload the data to.
    :raises common.ProtocolError: When the server does not comply to the tus protocol.
    """
    tus_headers = dict(headers or {})
    tus_headers["Tus-Resumable"] = common.TUS_PROTOCOL_VERSION

    if file is not None:
        total_size = await asyncio.to_thread(file.seek, 0, io.SEEK_END)
        tus_headers["Upload-Length"] = str(total_size)

    if metadata_header := encode_metadata(metadata):
        tus_headers["Upload-Metadata"] = metadata_header

    logger.debug("Creating upload...")
    async with await session.post(url, headers=tus_headers, ssl=ssl) as response:
        response.raise_for_status()
        if response.status != 201:  # noqa: PLR2004
            msg = f"Wrong status code {response.status}, expected 201."
            raise common.ProtocolError(msg)

        if "Location" not in response.headers:
            msg = 'Upload created, but no "Location" header in response.'
            raise common.ProtocolError(msg)

        return yarl.URL(response.headers["Location"])



"""Top-level module containing convenience functions."""

from __future__ import annotations

from .common import Metadata, ProtocolError, SSLArgument
from .retry import RetryConfiguration, metadata, upload, upload_multiple

__all__ = (
    "Metadata",
    "ProtocolError",
    "RetryConfiguration",
    "SSLArgument",
    "metadata",
    "upload",
    "upload_multiple",
)



"""Test the implementation of the core protocol."""

from __future__ import annotations

import binascii
from typing import TYPE_CHECKING

import aiohttp
import pytest

import aiotus

if TYPE_CHECKING:  # pragma: no cover
    import io

    import pytest_aiohttp

    from . import conftest


class TestOffset:
    async def test_offset_exceptions(
        self, aiohttp_server: pytest_aiohttp.AiohttpServer
    ) -> None:
        """Check for the different exceptions that can be thrown."""

        async def handler_not_found(_: aiohttp.web.Request) -> aiohttp.web.Response:
            raise aiohttp.web.HTTPNotFound

        async def handler_no_offset(_: aiohttp.web.Request) -> aiohttp.web.Response:
            raise aiohttp.web.HTTPOk

        async def handler_wrong_offset(_: aiohttp.web.Request) -> aiohttp.web.Response:
            headers = {"Upload-Offset": "xyz"}
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_negative_offset(
            _: aiohttp.web.Request,
        ) -> aiohttp.web.Response:
            headers = {"Upload-Offset": "-1"}
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("HEAD", "/not_found", handler_not_found)
        app.router.add_route("HEAD", "/no_offset", handler_no_offset)
        app.router.add_route("HEAD", "/wrong_offset", handler_wrong_offset)
        app.router.add_route("HEAD", "/negative_offset", handler_negative_offset)
        server = await aiohttp_server(app)

        # Check if status code is checked correctly.
        location = server.make_url("/not_found")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiohttp.ClientResponseError):
                await aiotus.core.offset(session, location)

        # Check if the check for the "Upload-Offset" header is working.
        location = server.make_url("/no_offset")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(
                aiotus.ProtocolError,
                match=r'HTTP header "Upload-Offset" not included in server response.',
            ):
                await aiotus.core.offset(session, location)

        # Check if the offset value is checked properly.
        location = server.make_url("/wrong_offset")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(
                aiotus.ProtocolError,
                match='Unable to convert "Upload-Offset" header',
            ):
                await aiotus.core.offset(session, location)

        location = server.make_url("/negative_offset")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(
                aiotus.ProtocolError,
                match='Unable to convert "Upload-Offset" header',
            ):
                await aiotus.core.offset(session, location)

    async def test_offset_functional(
        self, aiohttp_server: pytest_aiohttp.AiohttpServer
    ) -> None:
        """Test the normal functionality of the '_offset' function."""

        async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
            assert "Tus-Resumable" in request.headers
            assert (
                request.headers["Tus-Resumable"] == aiotus.common.TUS_PROTOCOL_VERSION
            )

            headers = {"Upload-Offset": "123"}
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("HEAD", "/files/12345678", handler)
        server = await aiohttp_server(app)

        location = server.make_url("/files/12345678")

        async with aiohttp.ClientSession() as session:
            offset = await aiotus.core.offset(session, location)

        assert offset == 123


class TestMetadata:
    def test_parse_metadata(self) -> None:
        """Check if metadata is parsed correctly."""

        md = aiotus.core._parse_metadata("")
        assert md == {}

        md = aiotus.core._parse_metadata("key")
        assert md == {"key": None}

        md = aiotus.core._parse_metadata("key ")
        assert md == {"key": None}

        md = aiotus.core._parse_metadata("key dmFsdWU=")
        assert md == {"key": b"value"}

        md = aiotus.core._parse_metadata("k1, k2 dmFsdWU=")
        assert md == {"k1": None, "k2": b"value"}

        md = aiotus.core._parse_metadata("k1 djE=, k2 djI=  ")
        assert md == {"k1": b"v1", "k2": b"v2"}

        with pytest.raises(binascii.Error, match="padding") as excinfo:
            aiotus.core._parse_metadata("k1 djE")

        with pytest.raises(binascii.Error) as excinfo:
            aiotus.core._parse_metadata("k1 dj&=")
        assert any(s in str(excinfo.value) for s in ("Non-base64", "Only base64"))

        with pytest.raises(ValueError, match="more than two elements"):
            aiotus.core._parse_metadata("k v v")

    async def test_metadata(self, aiohttp_server: pytest_aiohttp.AiohttpServer) -> None:
        """Check the 'core.metadata()' function."""

        async def handler_no_metadata(_: aiohttp.web.Request) -> aiohttp.web.Response:
            headers = {"Tus-Resumable": aiotus.common.TUS_PROTOCOL_VERSION}
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_invalid(_: aiohttp.web.Request) -> aiohttp.web.Response:
            headers = {
                "Tus-Resumable": aiotus.common.TUS_PROTOCOL_VERSION,
                "Upload-Metadata": "k1 djE",
            }
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_valid(_: aiohttp.web.Request) -> aiohttp.web.Response:
            headers = {
                "Tus-Resumable": aiotus.common.TUS_PROTOCOL_VERSION,
                "Upload-Metadata": "k1, k2 dmFsdWU=",
            }
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("HEAD", "/no_metadata", handler_no_metadata)
        app.router.add_route("HEAD", "/invalid", handler_invalid)
        app.router.add_route("HEAD", "/valid", handler_valid)
        server = await aiohttp_server(app)

        async with aiohttp.ClientSession() as session:
            md = await aiotus.core.metadata(session, server.make_url("/no_metadata"))
            assert md == {}

            with pytest.raises(aiotus.ProtocolError, match="Unable to parse metadata"):
                await aiotus.core.metadata(session, server.make_url("/invalid"))

            md = await aiotus.core.metadata(session, server.make_url("/valid"))
            assert md == {"k1": None, "k2": b"value"}


class TestUploadBuffer:
    async def test_upload_buffer(
        self,
        tus_server: conftest.MockTusServer,
        memory_file: io.BytesIO,
    ) -> None:
        """Test the upload function without simulating errors."""

        # We don't want to test the creation functionality here, so just create the
        # byte array so that 'tus_server' can accept data.
        tus_server.data = bytearray()

        async with aiohttp.ClientSession() as s:
            # Pass a small chunksize so that we have multiple uploads even with
            # the small test file.
            await aiotus.core.upload_buffer(
                s, tus_server.upload_endpoint, memory_file, ssl=False, chunksize=3
            )

        assert tus_server.data is not None
        assert tus_server.data == memory_file.getbuffer()

    async def test_server_offset(
        self,
        tus_server: conftest.MockTusServer,
        memory_file: io.BytesIO,
    ) -> None:
        """Test if the upload routine honors the offset value at the server-side."""

        tus_server.data = bytearray()

        # Make the server drop half of the uploaded data for each request.
        tus_server.drop_upload = True

        async with aiohttp.ClientSession() as s:
            await aiotus.core.upload_buffer(
                s, tus_server.upload_endpoint, memory_file, ssl=False, chunksize=3
            )

        assert tus_server.data is not None
        assert tus_server.data == memory_file.getbuffer()

    async def test_server_error(
        self,
        tus_server: conftest.MockTusServer,
        memory_file: io.BytesIO,
    ) -> None:
        """Simulate a server error."""

        tus_server.data = bytearray()
        tus_server.retries_upload = 2

        with pytest.raises(aiohttp.ClientResponseError) as excinfo:
            async with aiohttp.ClientSession() as s:
                await aiotus.core.upload_buffer(
                    s, tus_server.upload_endpoint, memory_file
                )

        assert "Internal Server Error" in str(excinfo.value)

    async def test_server_wrong_size(
        self,
        tus_server: conftest.MockTusServer,
        memory_file: io.BytesIO,
    ) -> None:
        """Simulate the case where the server has a bitter file than locally."""

        tus_server.data = bytearray(20 * b"x")

        with pytest.raises(aiotus.common.ProtocolError) as excinfo:
            async with aiohttp.ClientSession() as s:
                await aiotus.core.upload_buffer(
                    s, tus_server.upload_endpoint, memory_file
                )

        assert "Server offset too big" in str(excinfo.value)

    async def test_eof(
        self,
        tus_server: conftest.MockTusServer,
        eof_memory_file: io.BytesIO,
    ) -> None:
        """Simulate the case where the local file shrinks during an upload."""

        tus_server.data = bytearray()

        with pytest.raises(RuntimeError) as excinfo:
            async with aiohttp.ClientSession() as s:
                await aiotus.core.upload_buffer(
                    s, tus_server.upload_endpoint, eof_memory_file
                )

        assert "Buffer returned unexpected EOF" in str(excinfo.value)


class TestConfiguration:
    async def test_configuration_exceptions(
        self, aiohttp_server: pytest_aiohttp.AiohttpServer
    ) -> None:
        """Check for the different exceptions that can be thrown."""

        async def handler_not_found(_: aiohttp.web.Request) -> aiohttp.web.Response:
            raise aiohttp.web.HTTPNotFound

        async def handler_no_version(_: aiohttp.web.Request) -> aiohttp.web.Response:
            raise aiohttp.web.HTTPOk

        async def handler_max_size_invalid(
            _: aiohttp.web.Request,
        ) -> aiohttp.web.Response:
            headers = {"Tus-Version": "1.0.0", "Tus-Max-Size": "xyz"}
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_max_size_negative(
            _: aiohttp.web.Request,
        ) -> aiohttp.web.Response:
            headers = {"Tus-Version": "1.0.0", "Tus-Max-Size": "-1"}
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("OPTIONS", "/not_found", handler_not_found)
        app.router.add_route("OPTIONS", "/no_version", handler_no_version)
        app.router.add_route("OPTIONS", "/max_size_invalid", handler_max_size_invalid)
        app.router.add_route("OPTIONS", "/max_size_negative", handler_max_size_negative)
        server = await aiohttp_server(app)

        url = server.make_url("/not_found")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiohttp.ClientResponseError):
                await aiotus.core.configuration(session, url)

        url = server.make_url("/no_version")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiotus.ProtocolError):
                await aiotus.core.configuration(session, url)

        url = server.make_url("/max_size_invalid")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiotus.ProtocolError):
                await aiotus.core.configuration(session, url)

        url = server.make_url("/max_size_negative")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiotus.ProtocolError):
                await aiotus.core.configuration(session, url)

    async def test_configuration_function(
        self, aiohttp_server: pytest_aiohttp.AiohttpServer
    ) -> None:
        """Test the normal functionality of the 'configuration()' function."""

        async def handler_only_version(_: aiohttp.web.Request) -> aiohttp.web.Response:
            headers = {"Tus-Version": "1.0.0"}
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_all(_: aiohttp.web.Request) -> aiohttp.web.Response:
            headers = {
                "Tus-Version": "1.0.0,0.9.9",
                "Tus-Max-Size": "1024",
                "Tus-Extension": "creation,checksum",
            }
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("OPTIONS", "/only_version", handler_only_version)
        app.router.add_route("OPTIONS", "/all", handler_all)
        server = await aiohttp_server(app)

        # The server only returns the protocol version.
        url = server.make_url("/only_version")
        async with aiohttp.ClientSession() as session:
            config = await aiotus.core.configuration(session, url)

        assert len(config.protocol_versions) == 1
        assert config.protocol_versions[0] == "1.0.0"
        assert config.max_size is None
        assert len(config.protocol_extensions) == 0

        # The server also returns max size and protocol extensions.
        url = server.make_url("/all")
        async with aiohttp.ClientSession() as session:
            config = await aiotus.core.configuration(session, url)

        assert len(config.protocol_versions) == 2
        assert config.protocol_versions[0] == "1.0.0"
        assert config.protocol_versions[1] == "0.9.9"
        assert config.max_size == 1024
        assert len(config.protocol_extensions) == 2
        assert config.protocol_extensions[0] == "creation"
        assert config.protocol_extensions[1] == "checksum"

    async def test_configuration_function_tusd(self, tusd: conftest.TusServer) -> None:
        """Test the normal functionality of the 'configuration()' function with tusd."""

        async with aiohttp.ClientSession() as session:
            config = await aiotus.core.configuration(session, tusd.url)

        assert config.protocol_versions == ["1.0.0"]
        assert config.max_size is None
        assert "creation" in config.protocol_extensions



"""High-level functions that do retrying in case of communication errors."""

from __future__ import annotations

import asyncio
import contextlib
import dataclasses
from typing import TYPE_CHECKING

import aiohttp
import tenacity
import yarl

from . import common, core, creation
from .log import logger

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Iterable, Mapping
    from contextlib import AbstractAsyncContextManager
    from typing import BinaryIO


@dataclasses.dataclass
class RetryConfiguration:
    """Class to hold settings for the functions of this module."""

    retry_attempts: int = 10
    """
    Number of retry attempts to do when the communication fails.
    """

    max_retry_period_seconds: float = 60.0
    """
    Maximum time between retries, in seconds.

    Exponential backoff is used in case of communication errors,
    but the time between retries is caped by this value.
    """

    ssl: common.SSLArgument = True
    """
    'ssl' argument passed on to the aiohttp calls.

    This can be a boolean, or an instance of ssl.SSLContext, see
    the `aiohttp documentation
    <https://docs.aiohttp.org/en/stable/client_advanced.html#ssl-control-for-tcp-sockets>`_
    for the different meanings.
    """


def _make_log_before_function(s: str) -> Callable[[tenacity.RetryCallState], None]:
    """Create a function used to log before a retry attempt."""

    def log(retry_state: tenacity.RetryCallState) -> None:
        if retry_state.attempt_number > 1:
            logger.info(
                f"Trying {s} again, attempt number {retry_state.attempt_number}..."
            )

    return log


def _make_log_before_sleep_function(
    s: str,
) -> Callable[[tenacity.RetryCallState], None]:
    """Create a function used when a call made through tenacity fails."""

    def log(retry_state: tenacity.RetryCallState) -> None:
        if (retry_state.next_action is not None) and (retry_state.outcome is not None):
            duration = retry_state.next_action.sleep
            if retry_state.outcome.failed:
                value = retry_state.outcome.exception()
            else:
                value = retry_state.outcome.result()
            logger.warning(
                f"{s.capitalize()} failed, "
                f"retrying in {duration:.0f} second(s): {value}"
            )

    return log


def _make_retrying(s: str, config: RetryConfiguration) -> tenacity.AsyncRetrying:
    """Create a tenacity retry object."""
    return tenacity.AsyncRetrying(
        retry=tenacity.retry_if_exception_type(aiohttp.ClientError),
        stop=tenacity.stop_after_attempt(config.retry_attempts),
        wait=tenacity.wait_exponential(max=config.max_retry_period_seconds),
        before=_make_log_before_function(s),
        before_sleep=_make_log_before_sleep_function(s),
    )


def _sanitize_metadata(metadata: common.Metadata | None) -> common.Metadata:
    """Make sure the given optional metadata object is valid."""
    if metadata is None:
        metadata = {}

    # Try to encode the metadata once here, as this function is called outside
    # of any code that does retries, so that invalid arguments are caught early
    # and the exception is not swallowed.
    creation.encode_metadata(metadata)

    return metadata


async def upload(  # noqa: PLR0913
    endpoint: str | yarl.URL,
    file: BinaryIO,
    metadata: common.Metadata | None = None,
    client_session: aiohttp.ClientSession | None = None,
    config: RetryConfiguration | None = None,
    headers: Mapping[str, str] | None = None,
    chunksize: int = 4 * 1024 * 1024,
) -> yarl.URL | None:
    """Upload a file to a tus server.

    This function creates an upload on the server and then uploads
    the data to that location.

    In case of a communication error, this function retries the upload.

    :param endpoint: The creation endpoint of the server.
    :param file: The file to upload.
    :param metadata: Additional metadata for the upload.
    :param client_session: An aiohttp ClientSession to use.
    :param config: Settings to customize the retry behaviour.
    :param headers: Optional headers used in the request.
    :param chunksize: The size of individual chunks to upload at a time.
    :return: The location where the file was uploaded to (if the upload succeeded).
    """
    if config is None:
        config = RetryConfiguration()

    url = yarl.URL(endpoint)
    metadata = _sanitize_metadata(metadata)

    retrying_create = _make_retrying("upload creation", config)
    retrying_upload_file = _make_retrying("upload", config)

    try:
        ctx: aiohttp.ClientSession | AbstractAsyncContextManager[aiohttp.ClientSession]
        if client_session is None:
            ctx = aiohttp.ClientSession()
        else:
            ctx = contextlib.nullcontext(client_session)

        async with ctx as session:
            async for attempt in retrying_create:
                with attempt:
                    location = await creation.create(
                        session,
                        url,
                        file,
                        metadata,
                        ssl=config.ssl,
                        headers=headers,
                    )

            logger.debug(f"Upload created, upload URL is '{location}'.")
            if not location.is_absolute():
                location = url / location.path
                logger.debug(f"Upload URL was relative, changed to '{location}'.")

            async for attempt in retrying_upload_file:
                with attempt:
                    await core.upload_buffer(
                        session,
                        location,
                        file,
                        ssl=config.ssl,
                        chunksize=chunksize,
                        headers=headers,
                    )

            return location
    except tenacity.RetryError as e:
        logger.error(
            f"Unable to upload file, even after retrying: {e.last_attempt.exception()}"
        )
    except Exception as e:  # noqa: BLE001
        logger.error(f"Unable to upload file: {e}")

    return None


async def metadata(
    endpoint: str | yarl.URL,
    client_session: aiohttp.ClientSession | None = None,
    config: RetryConfiguration | None = None,
    headers: Mapping[str, str] | None = None,
) -> common.Metadata | None:
    """Read back the metadata of an upload.

    See :data:`aiotus.Metadata` for details on how metadata is handled in the
    tus protocol.

    In case of a communication error, this function retries.

    :param endpoint: The location of the upload.
    :param client_session: An aiohttp ClientSession to use.
    :param config: Settings to customize the retry behaviour.
    :param headers: Optional headers used in the request.
    :return: The metadata associated with the upload.
    """
    if config is None:
        config = RetryConfiguration()

    url = yarl.URL(endpoint)

    retrying_metadata = _make_retrying("query metadata", config)

    try:
        ctx: aiohttp.ClientSession | AbstractAsyncContextManager[aiohttp.ClientSession]
        if client_session is None:
            ctx = aiohttp.ClientSession()
        else:
            ctx = contextlib.nullcontext(client_session)

        async with ctx as session:
            async for attempt in retrying_metadata:
                with attempt:
                    return await core.metadata(
                        session, url, ssl=config.ssl, headers=headers
                    )
    except tenacity.RetryError as e:
        logger.error(
            f"Unable to get metadata, even after retrying: {e.last_attempt.exception()}"
        )

    return None


async def _upload_partial(  # noqa: PLR0913
    semaphore: asyncio.Semaphore,
    endpoint: str | yarl.URL,
    file: BinaryIO,
    client_session: aiohttp.ClientSession | None,
    config: RetryConfiguration,
    headers: Mapping[str, str] | None,
    chunksize: int,
) -> str:
    """Upload a single part of an upload with the "concatenation" extension.

    Helper function for 'upload_multiple()'.
    """
    tus_headers = dict(headers or {})
    tus_headers["Upload-Concat"] = "partial"

    async with semaphore:
        url = await upload(
            endpoint, file, None, client_session, config, tus_headers, chunksize
        )

    if url is None:
        msg = "Unable to upload part."
        raise RuntimeError(msg)

    return url.path


async def upload_multiple(  # noqa: C901 PLR0913
    endpoint: str | yarl.URL,
    files: Iterable[BinaryIO],
    metadata: common.Metadata | None = None,
    client_session: aiohttp.ClientSession | None = None,
    config: RetryConfiguration | None = None,
    headers: Mapping[str, str] | None = None,
    chunksize: int = 4 * 1024 * 1024,
    parallel_uploads: int = 3,
) -> yarl.URL | None:
    """Upload multiple files using the "concatenation" extension.

    Upload multiple files and then use the "concatenation" protocol extension
    to combine the parts on the server-side.

    :param endpoint: The creation endpoint of the server.
    :param files: The files to upload.
    :param metadata: Additional metadata for the final upload.
    :param client_session: An aiohttp ClientSession to use.
    :param config: Settings to customize the retry behaviour.
    :param headers: Optional headers used in the request.
    :param chunksize: The size of individual chunks to upload at a time.
    :param parallel_uploads: The number of parallel uploads to do concurrently.
    :return: The location of the final (concatenated) file on the server.
    :raises RuntimeError: If the server does not support the "concatenation" extension.
    """
    if config is None:
        config = RetryConfiguration()

    url = yarl.URL(endpoint)
    metadata = _sanitize_metadata(metadata)

    retrying_config = _make_retrying("query configuration", config)
    retrying_create = _make_retrying("upload creation", config)

    try:
        ctx: aiohttp.ClientSession | AbstractAsyncContextManager[aiohttp.ClientSession]
        if client_session is None:
            ctx = aiohttp.ClientSession()
        else:
            ctx = contextlib.nullcontext(client_session)

        async with ctx as session:
            #
            # Check if the server supports the "concatenation" extension.
            #
            async for attempt in retrying_config:
                with attempt:
                    server_config = await core.configuration(
                        session, url, ssl=config.ssl, headers=headers
                    )

            if "concatenation" not in server_config.protocol_extensions:
                msg = 'Server does not support the "concatenation" extension.'
                raise RuntimeError(msg)  # noqa: TRY301

            #
            # Upload the individual parts.
            #

            # Used to limit the number of coroutines that perform uploads in parallel.
            semaphore = asyncio.Semaphore(parallel_uploads)

            coros = [
                _upload_partial(
                    semaphore, endpoint, f, session, config, headers, chunksize
                )
                for f in files
            ]
            tasks = [asyncio.create_task(c) for c in coros]

            try:
                paths = await asyncio.gather(*tasks)
            except Exception as e:
                logger.info("Cancelling other uploads...")
                for t in tasks:
                    if not t.done():
                        t.cancel()

                msg = f"Upload of a part failed: {e}"
                raise RuntimeError(msg) from e

            concat_header = "final;" + " ".join(paths)

            #
            # Do the final concatenation.
            #
            final_headers = dict(headers or {})
            final_headers.update({"Upload-Concat": concat_header})

            async for attempt in retrying_create:  # pragma: no branch
                with attempt:
                    return await creation.create(
                        session,
                        url,
                        None,
                        metadata,
                        ssl=config.ssl,
                        headers=final_headers,
                    )
    except tenacity.RetryError as e:
        logger.error(
            f"Unable to upload files, even after retrying: {e.last_attempt.exception()}"
        )
    except Exception as e:  # noqa: BLE001
        logger.error(f"Unable to upload files: {e}")

    return None



"""Test the 'upload()' function."""

from __future__ import annotations

import io
import logging
from typing import TYPE_CHECKING

import aiohttp
import pytest
import tenacity
import yarl

import aiotus

if TYPE_CHECKING:  # pragma: no cover
    from . import conftest


class TestRetry:
    async def test_upload_functional(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Test the normal functionality of the 'upload()' function."""

        metadata = {"Content-Type": b"image/jpeg", "key": None}
        additional_headers = {"h1": "v1", "h2": "v2"}

        location = await aiotus.upload(
            tus_server.create_endpoint,
            memory_file,
            metadata,
            headers=additional_headers,
        )

        assert location is not None
        assert tus_server.metadata == "Content-Type aW1hZ2UvanBlZw==,key"
        assert tus_server.data is not None
        assert tus_server.data == memory_file.getbuffer()

        assert tus_server.post_headers is not None
        assert "h1" in tus_server.post_headers
        assert tus_server.post_headers["h1"] == "v1"
        assert "h2" in tus_server.post_headers
        assert tus_server.post_headers["h2"] == "v2"

        assert tus_server.head_headers is not None
        assert "h1" in tus_server.head_headers
        assert tus_server.head_headers["h1"] == "v1"
        assert "h2" in tus_server.head_headers
        assert tus_server.head_headers["h2"] == "v2"

    async def test_upload_client_session(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Use a custom client session."""

        headers = {"Authorization": "Basic xyz"}
        md1 = {"key1": b"value1", "key2": b"value2"}

        async with aiohttp.ClientSession(headers=headers) as s:
            location = await aiotus.upload(
                tus_server.create_endpoint, memory_file, md1, client_session=s
            )
            assert location is not None

            additional_headers = {"h1": "v1", "h2": "v2"}
            md2 = await aiotus.metadata(
                location, client_session=s, headers=additional_headers
            )

            assert not s.closed

        assert tus_server.data is not None
        assert tus_server.data == memory_file.getbuffer()
        assert tus_server.post_headers is not None
        assert "Authorization" in tus_server.post_headers
        assert tus_server.post_headers["Authorization"] == headers["Authorization"]
        assert tus_server.head_headers is not None
        assert "h1" in tus_server.head_headers
        assert tus_server.head_headers["h1"] == "v1"
        assert "h2" in tus_server.head_headers
        assert tus_server.head_headers["h2"] == "v2"
        assert md1 == md2

    async def test_upload_wrong_metadata(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Test if wrong metadata is rejected."""

        with pytest.raises(TypeError, match="bytes-like object is required"):
            await aiotus.upload(
                tus_server.create_endpoint,
                memory_file,
                {"k": "v"},  # type: ignore[dict-item]
            )

        with pytest.raises(ValueError, match="ASCII characters"):
            await aiotus.upload(tus_server.create_endpoint, memory_file, {"²": b"2"})

    async def test_retry(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Test the retry functionality."""

        # Make the server fail a few times to test the retry logic.
        tus_server.retries_head = 3
        tus_server.retries_upload = 3

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)
        md1 = {"key1": b"value1", "key2": b"value2"}

        location = await aiotus.upload(
            tus_server.create_endpoint, memory_file, md1, config=config
        )

        assert location is not None
        assert tus_server.data is not None
        assert tus_server.data == memory_file.getbuffer()

        tus_server.retries_head = 3
        md2 = await aiotus.metadata(location, config=config)

        assert md1 == md2

        tus_server.retries_head = 11
        md = await aiotus.metadata(location, config=config)

        assert md is None

    async def test_upload_create_retries_exceeded(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Not enough retries to create the upload."""

        tus_server.retries_create = 11

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)

        location = await aiotus.upload(
            tus_server.create_endpoint, memory_file, config=config
        )

        assert location is None
        assert tus_server.data is None  # Upload could not be created.

    async def test_upload_upload_retries_exceeded(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Not enough retries to do the upload."""

        tus_server.retries_upload = 11

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)

        location = await aiotus.upload(
            tus_server.create_endpoint, memory_file, config=config
        )

        assert location is None
        assert tus_server.data is not None  # Upload could be created.

    async def test_upload_relative_create(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Test what happens if the server returns a relative URL on creation."""

        # Make the create-handler return only the last part of the URL.
        path = tus_server.upload_endpoint.path
        tus_server.upload_endpoint = yarl.URL(path.split("/")[-1])

        location = await aiotus.upload(tus_server.create_endpoint, memory_file)

        assert location is not None
        assert tus_server.data is not None
        assert tus_server.data == memory_file.getbuffer()

    async def test_upload_file(self, tus_server: conftest.MockTusServer) -> None:
        """Test upload of an actual file, not an 'io.BytesIO'."""

        md1 = {"key1": b"value1", "key2": b"value2"}

        with open(__file__, "rb") as file:
            location = await aiotus.upload(tus_server.create_endpoint, file, md1)

        assert location is not None
        assert tus_server.data is not None
        with open(__file__, "rb") as file:
            data = file.read()
            assert tus_server.data == data

        md2 = await aiotus.metadata(str(location))
        assert md1 == md2

    async def test_tusd(
        self, tusd: conftest.TusServer, memory_file: io.BytesIO
    ) -> None:
        """Test communication with the the tusd server."""

        data = memory_file.getvalue()

        location = await aiotus.upload(tusd.url, memory_file)
        assert location is not None

        async with (
            aiohttp.ClientSession() as session,
            session.get(location) as response,
        ):
            body = await response.read()
            assert body == data

        md = await aiotus.metadata(str(location))
        assert not md  # We did not upload any metadata.

        # Upload a file with metadata to tusd, and check if we can read it back.
        md1 = {"key1": b"value1", "key2": b"value2"}
        location = await aiotus.upload(tusd.url, memory_file, md1)
        assert location is not None

        md2 = await aiotus.metadata(location)
        assert md1 == md2


class TestUploadMultiple:
    """Test the 'aiotus.upload_multiple()' function."""

    async def test_upload_functional(self, tusd: conftest.TusServer) -> None:
        """Upload files, read back as a single file."""

        file_a = io.BytesIO(b"\x00\x01")
        file_b = io.BytesIO(b"\x02\x03")
        file_c = io.BytesIO(b"\x04\x05")
        file_d = io.BytesIO(b"\x06\x07")

        md1 = {"key": b"value"}

        location = await aiotus.upload_multiple(
            tusd.url, [file_a, file_b, file_c, file_d], md1
        )
        assert location is not None

        md2 = await aiotus.metadata(location)
        assert md1 == md2

        async with (
            aiohttp.ClientSession() as session,
            session.get(location) as response,
        ):
            body = await response.read()
            assert body == b"\x00\x01\x02\x03\x04\x05\x06\x07"

    async def test_upload_functional_session(self, tusd: conftest.TusServer) -> None:
        """Upload files, read back as a single file, passing in a HTTP session."""

        file_a = io.BytesIO(b"\x00\x01")
        file_b = io.BytesIO(b"\x02\x03")
        file_c = io.BytesIO(b"\x04\x05")
        file_d = io.BytesIO(b"\x06\x07")

        async with aiohttp.ClientSession() as session:
            location = await aiotus.upload_multiple(
                tusd.url, [file_a, file_b, file_c, file_d], None, session
            )
            assert location is not None

            md = await aiotus.metadata(location)
            assert not md

            async with session.get(location) as response:
                body = await response.read()

                assert body == b"\x00\x01\x02\x03\x04\x05\x06\x07"

    async def test_not_supported(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Try uploading to a server that does not support concatenation."""

        location = await aiotus.upload_multiple(
            tus_server.create_endpoint, [memory_file]
        )
        assert location is None

    async def test_part_failure(self, tusd: conftest.TusServer) -> None:
        """Check the handling of a failure to upload a part."""

        file_a = io.BytesIO(b"\x00\x01")
        with open("/proc/self/mem", "rb") as file_b:
            location = await aiotus.upload_multiple(tusd.url, [file_a, file_b])

        assert location is None

    async def test_timeout(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Test handling of the retry exception."""

        tus_server.retries_options = 20

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)
        location = await aiotus.upload_multiple(
            tus_server.create_endpoint, [memory_file], config=config
        )

        assert location is None


class TestTenacity:
    """Test the log functions for tenacity."""

    async def test_exception_logging(self, caplog: pytest.LogCaptureFixture) -> None:
        rt = tenacity.AsyncRetrying(
            retry=tenacity.retry_if_exception_type(RuntimeError),
            stop=tenacity.stop_after_attempt(3),
            before=aiotus.retry._make_log_before_function("test"),
            before_sleep=aiotus.retry._make_log_before_sleep_function("test"),
        )

        with caplog.at_level(logging.INFO, logger="aiotus"):
            with pytest.raises(tenacity.RetryError):  # noqa: PT012
                async for attempt in rt:
                    with attempt:
                        await TestTenacity.raise_runtime_error()

            lg = caplog.record_tuples
            assert len(lg) == 4
            assert lg[0][2] == "Test failed, retrying in 0 second(s): test error"
            assert lg[1][2] == "Trying test again, attempt number 2..."
            assert lg[2][2] == "Test failed, retrying in 0 second(s): test error"
            assert lg[3][2] == "Trying test again, attempt number 3..."

    async def test_result_logging(self, caplog: pytest.LogCaptureFixture) -> None:
        rt = tenacity.AsyncRetrying(
            retry=tenacity.retry_if_result(lambda r: r is None),
            stop=tenacity.stop_after_attempt(2),
            before=aiotus.retry._make_log_before_function("test"),
            before_sleep=aiotus.retry._make_log_before_sleep_function("test"),
        )

        with caplog.at_level(logging.INFO, logger="aiotus"):
            with pytest.raises(tenacity.RetryError):  # noqa: PT012
                async for attempt in rt:
                    with attempt:
                        await TestTenacity.return_none()

            lg = caplog.record_tuples
            assert len(lg) == 2
            assert lg[0][2] == "Test failed, retrying in 0 second(s): None"
            assert lg[1][2] == "Trying test again, attempt number 2..."

    @staticmethod
    async def raise_runtime_error() -> None:
        msg = "test error"
        raise RuntimeError(msg)

    @staticmethod
    async def return_none() -> None:
        return None



"""Test the implementation of the creation extension."""

from __future__ import annotations

from typing import TYPE_CHECKING

import aiohttp
import pytest

import aiotus

if TYPE_CHECKING:  # pragma: no cover
    import io

    import pytest_aiohttp

    from . import conftest


class TestCreate:
    async def test_create_wrong_metadata(self, memory_file: io.BytesIO) -> None:
        """Check the different checks performed on metadata keys."""

        metadata = {"k1": b"v1", "k²": b"v2", "k3": b"v3"}
        with pytest.raises(ValueError, match="ASCII characters"):
            await aiotus.creation.create(
                None,  # type: ignore[arg-type]
                None,  # type: ignore[arg-type]
                memory_file,
                metadata,
            )

        metadata = {"k1": b"v1", "k 2": b"v2", "k3": b"v3"}
        with pytest.raises(ValueError, match="spaces"):
            await aiotus.creation.create(
                None,  # type: ignore[arg-type]
                None,  # type: ignore[arg-type]
                memory_file,
                metadata,
            )

        metadata = {"k1": b"v1", "k2,": b"v2", "k3": b"v3"}
        with pytest.raises(ValueError, match="commas"):
            await aiotus.creation.create(
                None,  # type: ignore[arg-type]
                None,  # type: ignore[arg-type]
                memory_file,
                metadata,
            )

    async def test_create_wrong_status(
        self, aiohttp_server: pytest_aiohttp.AiohttpServer, memory_file: io.BytesIO
    ) -> None:
        """Check if status code is checked correctly."""

        async def handler_status_200(
            _: aiohttp.web.Request,
        ) -> aiohttp.web.Response:
            return aiohttp.web.Response(status=200)

        async def handler_status_400(
            _: aiohttp.web.Request,
        ) -> aiohttp.web.Response:
            return aiohttp.web.Response(status=400)

        app = aiohttp.web.Application()
        app.router.add_route("POST", "/status_200", handler_status_200)
        app.router.add_route("POST", "/status_400", handler_status_400)
        server = await aiohttp_server(app)

        endpoint = server.make_url("/status_200")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiotus.ProtocolError, match="Wrong status code"):
                await aiotus.creation.create(session, endpoint, memory_file, {})

        endpoint = server.make_url("/status_400")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiohttp.ClientResponseError) as excinfo:
                await aiotus.creation.create(session, endpoint, memory_file, {})
        assert excinfo.value.status == 400

    async def test_create_no_location(
        self, aiohttp_server: pytest_aiohttp.AiohttpServer, memory_file: io.BytesIO
    ) -> None:
        """Check if the check for the "Location" header is working."""

        async def handler_no_location(
            _: aiohttp.web.Request,
        ) -> aiohttp.web.Response:
            raise aiohttp.web.HTTPCreated

        app = aiohttp.web.Application()
        app.router.add_route("POST", "/no_location", handler_no_location)
        server = await aiohttp_server(app)

        endpoint = server.make_url("/no_location")
        async with aiohttp.ClientSession() as session:
            with pytest.raises(aiotus.ProtocolError) as excinfo:
                await aiotus.creation.create(session, endpoint, memory_file, {})

        assert 'no "Location" header' in str(excinfo.value)

    async def test_create_functional(
        self, tus_server: conftest.MockTusServer, memory_file: io.BytesIO
    ) -> None:
        """Test the normal functionality of the upload creation."""

        metadata = {"k1": b"1", "k2": "2²".encode(), "k-3": b"three"}

        endpoint = tus_server.server.make_url("/files")

        async with aiohttp.ClientSession() as session:
            location = await aiotus.creation.create(
                session, endpoint, memory_file, metadata
            )

        assert tus_server.upload_endpoint == location



"""Test the entrypoint module."""

from __future__ import annotations

import os.path
import unittest.mock
from typing import TYPE_CHECKING

import aiotus.entrypoint

if TYPE_CHECKING:  # pragma: no cover
    import pytest

    from . import conftest


class TestAiotusClients:
    def test_no_command(self, capsys: pytest.CaptureFixture[str]) -> None:
        assert aiotus.entrypoint.main([]) == 1

        _, stderr = capsys.readouterr()
        lines = stderr.splitlines(keepends=False)
        assert len(lines) >= 0
        assert lines[0] == "No command specified."

    def test_aiotus_clients(
        self, capsys: pytest.CaptureFixture[str], tusd: conftest.TusServer
    ) -> None:
        conf = aiotus.RetryConfiguration(1, 0.001, ssl=False)
        defaults = (None, None, conf, None, 4 * 1024 * 1024)
        argv = ["--debug", "upload", str(tusd.url) + "x", __file__]
        with unittest.mock.patch.object(aiotus.upload, "__defaults__", defaults):
            assert aiotus.entrypoint.main(argv) == 1

        argv = ["upload", str(tusd.url), __file__ + "x"]
        assert aiotus.entrypoint.main(argv) == 1

        argv = ["upload", str(tusd.url), __file__]
        assert aiotus.entrypoint.main(argv) == 0

        stdout, _ = capsys.readouterr()
        url = stdout.splitlines(keepends=False)[0]

        expected_output = [
            f"filename: {os.path.basename(__file__)}",
            "mime_type: text/x-python",
        ]

        argv = ["metadata", url]
        assert aiotus.entrypoint.main(argv) == 0

        stdout, _ = capsys.readouterr()
        lines = stdout.splitlines(keepends=False)
        lines.sort()

        assert lines == expected_output

        argv = ["--debug", "metadata", url]
        assert aiotus.entrypoint.main(argv) == 0

        stdout, _ = capsys.readouterr()
        lines = stdout.splitlines(keepends=False)
        lines.sort()
        assert lines == expected_output

    def test_additional_metadata(
        self, capsys: pytest.CaptureFixture[str], tusd: conftest.TusServer
    ) -> None:
        argv = [
            "upload",
            "--metadata",
            "key1=value1",
            "--metadata",
            "key2",
            str(tusd.url),
            __file__,
        ]
        assert aiotus.entrypoint.main(argv) == 0

        stdout, _ = capsys.readouterr()
        url = stdout.splitlines(keepends=False)[0]

        argv = ["metadata", url]
        assert aiotus.entrypoint.main(argv) == 0

        stdout, _ = capsys.readouterr()
        lines = stdout.splitlines(keepends=False)
        assert "key1: value1" in lines
        assert "key2" in lines

    def test_no_metadata(self) -> None:
        argv = ["metadata", "http://example.com"]
        with unittest.mock.patch("aiotus.retry.metadata", return_value=None):
            assert aiotus.entrypoint.main(argv) == 1
